from typing import Optional, Set, Any

from fakeredis.commands_mixins import (
    ArrayCommandsMixin,
    BitmapCommandsMixin,
    ConnectionCommandsMixin,
    GenericCommandsMixin,
    GeoCommandsMixin,
    HashCommandsMixin,
    ListCommandsMixin,
    PubSubCommandsMixin,
    ScriptingCommandsMixin,
    ServerCommandsMixin,
    StringCommandsMixin,
    TransactionsCommandsMixin,
    SetCommandsMixin,
    StreamsCommandsMixin,
    AclCommandsMixin,
)
from fakeredis.stack import (
    JSONCommandsMixin,
    BFCommandsMixin,
    CFCommandsMixin,
    CMSCommandsMixin,
    TopkCommandsMixin,
    TDigestCommandsMixin,
    TimeSeriesCommandsMixin,
    VectorSetCommandsMixin,
)
from ._basefakesocket import BaseFakeSocket
from ._server import FakeServer
from .commands_mixins.sortedset_mixin import SortedSetCommandsMixin
from .server_specific_commands import DragonflyCommandsMixin


class FakeSocket(
    BaseFakeSocket,
    ArrayCommandsMixin,
    GenericCommandsMixin,
    ScriptingCommandsMixin,
    HashCommandsMixin,
    ConnectionCommandsMixin,
    ListCommandsMixin,
    ServerCommandsMixin,
    StringCommandsMixin,
    TransactionsCommandsMixin,
    PubSubCommandsMixin,
    SetCommandsMixin,
    BitmapCommandsMixin,
    SortedSetCommandsMixin,
    StreamsCommandsMixin,
    JSONCommandsMixin,
    GeoCommandsMixin,
    BFCommandsMixin,
    CFCommandsMixin,
    CMSCommandsMixin,
    TopkCommandsMixin,
    TDigestCommandsMixin,
    TimeSeriesCommandsMixin,
    DragonflyCommandsMixin,
    AclCommandsMixin,
    VectorSetCommandsMixin,
):
    def __init__(
        self,
        server: FakeServer,
        db: int,
        lua_modules: Optional[Set[str]] = None,  # noqa: F821
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super(FakeSocket, self).__init__(server, db, *args, lua_modules=lua_modules, **kwargs)



"""Helpers for constructing the underlying redis/valkey client.

Used by the sync and async ``FakeRedisMixin`` classes to translate the
arguments accepted by ``FakeRedis(...)`` (and friends) into the kwargs the
real client class expects, wiring in a fakeredis connection pool.
"""

import inspect
import uuid
import warnings
from typing import Any, Callable, Dict, Optional, Set, Type


def _get_args_to_warn(method: Callable[..., Any]) -> Set[str]:
    """Collect argument names that ``method`` would emit deprecation warnings for.

    redis-py marks deprecated ``__init__`` arguments by wrapping the method in
    a ``@deprecated_args(args_to_warn=[...])`` decorator. There is no public
    API to query the list, so this walks the wrapper's closure cells looking
    for the ``args_to_warn`` list (recursing through nested wrappers). If
    redis-py changes how the decorator stores the list, this returns an empty
    set and deprecated defaults are simply forwarded again.
    """
    closure = method.__closure__
    if closure is None:
        return set()
    res = set()
    for cell in closure:
        value = cell.cell_contents
        if isinstance(value, list) and len(value) > 0:
            res.update(value)
        elif callable(value):
            res.update(_get_args_to_warn(value))
    return res


def convert_args_kwargs(klass: Type[object], *args: Any, **kwargs: Any) -> Dict[str, Any]:
    """Interpret the positional and keyword arguments according to the version of redis in use"""
    parameters = list(inspect.signature(klass.__init__).parameters.values())[1:]
    args_to_warn = _get_args_to_warn(klass.__init__)
    # Convert args => kwargs
    kwargs.update({parameters[i].name: args[i] for i in range(len(args))})
    if "path" not in kwargs and "host" not in kwargs:
        kwargs["host"] = uuid.uuid4().hex
    kwds = {
        p.name: kwargs.get(p.name, p.default)
        for ind, p in enumerate(parameters)
        if p.default != inspect.Parameter.empty and (p.name not in args_to_warn or p.name in kwargs)
    }
    return kwds


# Client kwargs that are forwarded to the connection pool / connection.
_CONNECTION_POOL_KWARGS = frozenset(
    {
        "host",
        "port",
        "db",
        "username",
        "password",
        "socket_timeout",
        "encoding",
        "encoding_errors",
        "decode_responses",
        "retry_on_timeout",
        "max_connections",
        "health_check_interval",
        "client_name",
        "connected",
        "server",
        "protocol",
    }
)


def build_client_kwds(
    *args: Any,
    client_class: Type[Any],
    connection_class: Type[Any],
    connection_pool_class: Type[Any],
    version: Any,
    server_type: Any,
    lua_modules: Any,
    server: Any,
    connected: Optional[bool] = None,
    **kwargs: Any,
) -> Dict[str, Any]:
    """Build the kwargs passed to the underlying redis/valkey client ``__init__``.

    Creates a fakeredis connection pool when one isn't supplied. Shared by the sync
    and async ``FakeRedisMixin`` classes; the caller still applies any lib_name /
    driver_info handling specific to its client library.
    """
    kwds = convert_args_kwargs(client_class, *args, **kwargs)
    kwds["server"] = server
    if connected is not None:
        kwds["connected"] = connected
    if not kwds.get("connection_pool", None):
        # Adapted from redis-py: translate the deprecated charset/errors aliases.
        charset = kwds.get("charset", None)
        if charset is not None:
            warnings.warn(DeprecationWarning('"charset" is deprecated. Use "encoding" instead'))
            kwds["encoding"] = charset
        errors = kwds.get("errors", None)
        if errors is not None:
            warnings.warn(DeprecationWarning('"errors" is deprecated. Use "encoding_errors" instead'))
            kwds["encoding_errors"] = errors
        connection_kwargs: Dict[str, Any] = {
            "connection_class": connection_class,
            "version": version,
            "server_type": server_type,
            "lua_modules": lua_modules,
            "client_class": client_class,
        }
        connection_kwargs.update({arg: kwds[arg] for arg in _CONNECTION_POOL_KWARGS if arg in kwds})
        kwds["connection_pool"] = connection_pool_class(**connection_kwargs)
    for key in ("server", "connected", "version", "server_type", "lua_modules"):
        kwds.pop(key, None)
    return kwds



from typing import Any, List, Optional, Tuple

from fakeredis import _msgs as msgs
from fakeredis._commands import Key, Int, command, CommandItem
from fakeredis._helpers import SimpleError, casematch
from fakeredis.commands_mixins._mixin_base import CommandsMixinBase
from fakeredis.model._array import Array

_ARRAY_WRONG_TYPE = msgs.WRONGTYPE_MSG
_INDEX_ERROR = "ERR index out of range"
_INVALID_SIZE = "ERR invalid size"
_UNKNOWN_OP = "ERR unknown operation"


def _parse_range_end(val: bytes, array: Array, is_start: bool) -> int:
    """Parse start/end arg: b'-' → 0, b'+' → max_index, otherwise int."""
    if val == b"-":
        return 0
    if val == b"+":
        return max(array._data.keys(), default=0)
    try:
        return int(val)
    except ValueError:
        raise SimpleError(msgs.INVALID_INT_MSG)


class ArrayCommandsMixin(CommandsMixinBase):
    # ── write commands ──────────────────────────────────────────────────────

    @command((Key(Array), Int, bytes), (bytes,))
    def arset(self, key: CommandItem, index: int, first_val: bytes, *more_vals: bytes) -> int:
        if index < 0:
            raise SimpleError(_INDEX_ERROR)
        arr: Array = key.value
        new_slots = 0
        for i, val in enumerate([first_val] + list(more_vals)):
            if arr.set(index + i, val):
                new_slots += 1
        key.updated()
        return new_slots

    @command((Key(Array), Int, bytes), (Int, bytes))
    def armset(self, key: CommandItem, *args: Any) -> int:
        # args: index1, val1, index2, val2, ...
        arr: Array = key.value
        new_slots = 0
        for i in range(0, len(args), 2):
            idx, val = args[i], args[i + 1]
            if idx < 0:
                raise SimpleError(_INDEX_ERROR)
            if arr.set(idx, val):
                new_slots += 1
        key.updated()
        return new_slots

    @command((Key(Array), bytes), (bytes,))
    def arinsert(self, key: CommandItem, first_val: bytes, *more_vals: bytes) -> int:
        arr: Array = key.value
        last_idx = arr._cursor
        for val in [first_val] + list(more_vals):
            arr.set(arr._cursor, val)
            arr.record_insert(arr._cursor)
            last_idx = arr._cursor
            arr._cursor += 1
        key.updated()
        return last_idx

    @command((Key(Array), Int, bytes), (bytes,))
    def arring(self, key: CommandItem, size: int, first_val: bytes, *more_vals: bytes) -> int:
        if size <= 0:
            raise SimpleError(_INVALID_SIZE)
        arr: Array = key.value
        if arr.length() > size:
            arr.truncate_at(size)
        last_idx = arr._cursor % size
        for val in [first_val] + list(more_vals):
            idx = arr._cursor % size
            arr.set(idx, val)
            arr.record_insert(idx)
            last_idx = idx
            arr._cursor += 1
        key.updated()
        return last_idx

    @command((Key(Array, 0), Int), (Int,))
    def ardel(self, key: CommandItem, first_idx: int, *more_idx: int) -> int:
        if not key:
            return 0
        arr: Array = key.value
        deleted = 0
        for idx in [first_idx] + list(more_idx):
            if arr.delete(idx):
                deleted += 1
        if deleted:
            key.updated()
        return deleted

    @command((Key(Array, 0), Int, Int), (Int, Int))
    def ardelrange(self, key: CommandItem, *args: int) -> int:
        if not key:
            return 0
        arr: Array = key.value
        # args: start1, end1, start2, end2, ...
        to_delete = set()
        for i in range(0, len(args), 2):
            start, end = args[i], args[i + 1]
            lo, hi = (start, end) if start <= end else (end, start)
            to_delete.update(k for k in arr._data if lo <= k <= hi)
        for idx in to_delete:
            arr.delete(idx)
        if to_delete:
            key.updated()
        return len(to_delete)

    @command((Key(Array, 0), Int))
    def arseek(self, key: CommandItem, index: int) -> int:
        if not key:
            return 0
        arr: Array = key.value
        arr._cursor = index
        key.updated()
        return 1

    # ── read commands ────────────────────────────────────────────────────────

    @command((Key(Array, None), Int))
    def arget(self, key: CommandItem, index: int) -> Optional[bytes]:
        if not key:
            return None
        return key.value.get(index)

    @command((Key(Array), Int, Int))
    def argetrange(self, key: CommandItem, start: int, end: int) -> List[Optional[bytes]]:
        if start <= end:
            indices = list(range(start, end + 1))
        else:
            indices = list(range(start, end - 1, -1))
        if not key:
            return [None] * len(indices)
        arr: Array = key.value
        return [arr.get(i) for i in indices]

    @command((Key(Array), Int), (Int,))
    def armget(self, key: CommandItem, first_idx: int, *more_idx: int) -> List[Optional[bytes]]:
        arr: Array = key.value
        return [arr.get(idx) for idx in [first_idx] + list(more_idx)]

    @command((Key(Array, 0),))
    def arlen(self, key: CommandItem) -> int:
        if not key:
            return 0
        return key.value.length()

    @command((Key(Array, 0),))
    def arcount(self, key: CommandItem) -> int:
        if not key:
            return 0
        return key.value.count()

    @command((Key(Array, 0),))
    def arnext(self, key: CommandItem) -> int:
        if not key:
            return 0
        return key.value._cursor

    @command((Key(Array, []), Int), (bytes,))
    def arlastitems(self, key: CommandItem, count: int, *args: bytes) -> List[bytes]:
        if not key:
            return []
        rev = any(casematch(a, b"rev") for a in args)
        items = key.value.lastitems(count)
        if rev:
            items = list(reversed(items))
        return items

    @command((Key(Array, []), bytes, bytes), (bytes,))
    def arscan(self, key: CommandItem, start_b: bytes, end_b: bytes, *args: bytes) -> List[Any]:
        if not key:
            return []
        arr: Array = key.value
        try:
            start = int(start_b)
            end = int(end_b)
        except ValueError:
            raise SimpleError(msgs.INVALID_INT_MSG)

        limit: Optional[int] = None
        i = 0
        while i < len(args):
            if casematch(args[i], b"limit"):
                if i + 1 >= len(args):
                    raise SimpleError(msgs.SYNTAX_ERROR_MSG)
                try:
                    limit = int(args[i + 1])
                except ValueError:
                    raise SimpleError(msgs.INVALID_INT_MSG)
                i += 2
            else:
                raise SimpleError(msgs.SYNTAX_ERROR_MSG)

        pairs = arr.scan_range(start, end, limit)
        return [[idx, val] for idx, val in pairs]

    @command((Key(Array, []), bytes, bytes), (bytes,))
    def argrep(self, key: CommandItem, start_b: bytes, end_b: bytes, *args: bytes) -> List[Any]:
        if not key:
            return []
        arr: Array = key.value
        start = _parse_range_end(start_b, arr, is_start=True)
        end = _parse_range_end(end_b, arr, is_start=False)

        predicates: List[Tuple[str, str]] = []
        use_and = False
        limit: Optional[int] = None
        withvalues = False
        nocase = False

        i = 0
        while i < len(args):
            a = args[i]
            if casematch(a, b"exact") or casematch(a, b"match") or casematch(a, b"glob") or casematch(a, b"re"):
                if i + 1 >= len(args):
                    raise SimpleError(msgs.SYNTAX_ERROR_MSG)
                kind = a.lower().decode()
                predicates.append((kind, args[i + 1].decode(errors="replace")))
                i += 2
            elif casematch(a, b"and"):
                use_and = True
                i += 1
            elif casematch(a, b"or"):
                use_and = False
                i += 1
            elif casematch(a, b"limit"):
                if i + 1 >= len(args):
                    raise SimpleError(msgs.SYNTAX_ERROR_MSG)
                try:
                    limit = int(args[i + 1])
                except ValueError:
                    raise SimpleError(msgs.INVALID_INT_MSG)
                i += 2
            elif casematch(a, b"withvalues"):
                withvalues = True
                i += 1
            elif casematch(a, b"nocase"):
                nocase = True
                i += 1
            else:
                raise SimpleError(msgs.SYNTAX_ERROR_MSG)

        if not predicates:
            raise SimpleError(msgs.SYNTAX_ERROR_MSG)

        matches = arr.grep_range(start, end, predicates, use_and, limit, nocase)
        result: List[Any] = []
        for idx, val in matches:
            if withvalues:
                result.append([idx, val])
            else:
                result.append(idx)
        return result

    @command((Key(Array, None), bytes, bytes, bytes), (bytes,))
    def arop(self, key: CommandItem, start_b: bytes, end_b: bytes, op_b: bytes, *args: bytes) -> Any:
        if not key:
            return None
        arr: Array = key.value
        try:
            start = int(start_b)
            end = int(end_b)
        except ValueError:
            raise SimpleError(msgs.INVALID_INT_MSG)

        op = op_b.lower().decode()
        operand: Optional[bytes] = None
        if op == "match":
            if not args:
                raise SimpleError(msgs.SYNTAX_ERROR_MSG)
            operand = args[0]
        elif op not in ("sum", "min", "max", "and", "or", "xor", "used"):
            raise SimpleError(_UNKNOWN_OP)

        return arr.op_range(start, end, op, operand)

    @command((Key(Array),), (bytes,))
    def arinfo(self, key: CommandItem, *args: bytes) -> Any:
        if not key:
            raise SimpleError("no such key")
        arr: Array = key.value
        full = any(casematch(a, b"full") for a in args)
        info: List[Any] = [
            b"count",
            arr.count(),
            b"len",
            arr.length(),
            b"next-insert-index",
            arr._cursor,
            b"slices",
            1,
            b"directory-size",
            1,
            b"super-dir-entries",
            0,
            b"slice-size",
            4096,
        ]
        if full:
            info += [
                b"dense-slices",
                0,
                b"sparse-slices",
                1,
                b"avg-dense-size",
                0.0,
                b"avg-dense-fill",
                0.0,
                b"avg-sparse-size",
                float(arr.count() * 4),
            ]
        return info



from sortedcontainers import SortedList

from ._base_type import BaseModel


class TDigest(SortedList, BaseModel):  # type: ignore[misc]
    _model_type = b"TDIS-TYPE"

    def __init__(self, compression: int = 100) -> None:
        super().__init__()
        self.compression = compression



from __future__ import annotations

from typing import TYPE_CHECKING

from fakeredis._typing import VersionType

if TYPE_CHECKING:
    from fakeredis._helpers import Database
    from fakeredis._server import FakeServer
    from fakeredis.model import ClientInfo


class CommandsMixinBase:
    """Base class for command mixins that declares shared read-only attributes."""

    _server: FakeServer
    _client_info: ClientInfo
    _db: Database

    @property
    def version(self) -> VersionType:
        raise NotImplementedError



import hashlib
import pickle
import random
from typing import Tuple, Any, Callable, List, Optional, Union, Sequence

from fakeredis import _msgs as msgs
from fakeredis._command_args_parsing import extract_args
from fakeredis._commands import (
    command,
    Key,
    Int,
    DbIndex,
    BeforeAny,
    CommandItem,
    SortFloat,
    delete_keys,
)
from fakeredis._helpers import compile_pattern, SimpleError, OK, casematch, SimpleString
from fakeredis.commands_mixins._mixin_base import CommandsMixinBase
from fakeredis.model import ZSet, Hash, ExpiringMembersSet


class GenericCommandsMixin(CommandsMixinBase):
    _ttl: Callable[[CommandItem, float], int]
    _scan: Callable[[Sequence[bytes], int, bytes], List[Union[bytes, List[bytes]]]]
    _key_value_type: Callable[[CommandItem], SimpleString]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(GenericCommandsMixin, self).__init__(*args, **kwargs)
        self._db_num: int

    def _lookup_key(self, key: bytes, pattern: bytes) -> Optional[bytes]:
        """Python implementation of lookupKeyByPattern from redis"""
        if pattern == b"#":
            return key
        p = pattern.find(b"*")
        if p == -1:
            return None
        prefix = pattern[:p]
        suffix = pattern[p + 1 :]
        arrow = suffix.find(b"->", 0, -1)
        if arrow != -1:
            field = suffix[arrow + 2 :]
            suffix = suffix[:arrow]
        else:
            field = None
        new_key = prefix + key + suffix
        item = CommandItem(new_key, self._db, item=self._db.get(new_key))
        if item.value is None:
            return None
        if field is not None:
            if not isinstance(item.value, Hash):
                return None
            return item.value.get(field)  # type: ignore
        else:
            if not isinstance(item.value, bytes):
                return None
            return item.value

    def _expireat(self, key: CommandItem, timestamp: float, *args: bytes) -> int:
        ((nx, xx, gt, lt), _) = extract_args(args, ("nx", "xx", "gt", "lt"), exception=msgs.EXPIRE_UNSUPPORTED_OPTION)
        if self.version < (7,) and any((nx, xx, gt, lt)):
            raise SimpleError(msgs.WRONG_ARGS_MSG6.format("expire"))
        counter = (nx, gt, lt).count(True)
        if (counter > 1) or (nx and xx):
            raise SimpleError(msgs.NX_XX_GT_LT_ERROR_MSG)
        if (
            not key
            or (xx and key.expireat is None)
            or (nx and key.expireat is not None)
            # A key with no expiry is treated as infinity: GT never sets it (nothing is greater than infinity) while
            # LT always does. GT/LT are also strict, so an equal timestamp must not set either.
            or (gt and (key.expireat is None or timestamp <= key.expireat))
            or (lt and key.expireat is not None and timestamp >= key.expireat)
        ):
            return 0
        key.expireat = timestamp
        return 1

    @command(name="DEL", fixed=(Key(),), repeat=(Key(),))
    def del_(self, *keys: CommandItem) -> int:
        return delete_keys(*keys)

    @command(name="DUMP", fixed=(Key(missing_return=None),))
    def dump(self, key: CommandItem) -> Optional[bytes]:
        value = pickle.dumps(key.value)
        checksum = hashlib.sha1(value).digest()
        return checksum + value

    @command(name="EXISTS", fixed=(Key(),), repeat=(Key(),))
    def exists(self, *keys: CommandItem) -> int:
        ret = 0
        for key in keys:
            if key:
                ret += 1
        return ret

    @command(name="EXPIRE", fixed=(Key(), Int), repeat=(bytes,))
    def expire(self, key: CommandItem, seconds: int, *args: bytes) -> int:
        res = self._expireat(key, self._db.time + seconds, *args)
        return res

    @command(name="EXPIREAT", fixed=(Key(), Int), repeat=(bytes,))
    def expireat(self, key: CommandItem, timestamp: int, *args: bytes) -> int:
        return self._expireat(key, float(timestamp), *args)

    @command(name="KEYS", fixed=(bytes,))
    def keys(self, pattern: bytes) -> List[bytes]:
        if pattern == b"*":
            return list(self._db)
        else:
            regex = compile_pattern(pattern)
            return [key for key in self._db if regex.match(key)]

    @command(name="MOVE", fixed=(Key(), DbIndex))
    def move(self, key: CommandItem, db: int) -> int:
        if db == self._db_num:
            raise SimpleError(msgs.SRC_DST_SAME_MSG)
        if not key or key.key in self._server.dbs[db]:
            return 0
        # TODO: what is the interaction with expiry?
        self._server.dbs[db][key.key] = self._server.dbs[self._db_num][key.key]
        key.value = None  # Causes deletion
        return 1

    @command(name="PERSIST", fixed=(Key(),))
    def persist(self, key: CommandItem) -> int:
        if key.expireat is None:
            return 0
        key.expireat = None
        return 1

    @command(name="PEXPIRE", fixed=(Key(), Int), repeat=(bytes,))
    def pexpire(self, key: CommandItem, ms: int, *args: bytes) -> int:
        return self._expireat(key, self._db.time + ms / 1000.0, *args)

    @command(name="PEXPIREAT", fixed=(Key(), Int), repeat=(bytes,))
    def pexpireat(self, key: CommandItem, ms_timestamp: int, *args: bytes) -> int:
        return self._expireat(key, ms_timestamp / 1000.0, *args)

    @command(name="PTTL", fixed=(Key(),))
    def pttl(self, key: CommandItem) -> int:
        return self._ttl(key, 1000.0)

    @command(name="EXPIRETIME", fixed=(Key(),))
    def expiretime(self, key: CommandItem) -> int:
        if key.value is None:
            return -2
        if key.expireat is None:
            return -1
        return int(key.expireat)

    @command(name="PEXPIRETIME", fixed=(Key(),))
    def pexpiretime(self, key: CommandItem) -> int:
        if key.value is None:
            return -2
        if key.expireat is None:
            return -1
        return int(key.expireat * 1000)

    @command(name="RANDOMKEY", fixed=())
    def randomkey(self) -> Optional[bytes]:
        keys: List[bytes] = list(self._db.keys())
        if not keys:
            return None
        return random.choice(keys)

    @command(name="RENAME", fixed=(Key(), Key()))
    def rename(self, key: CommandItem, newkey: CommandItem) -> SimpleString:
        if not key:
            raise SimpleError(msgs.NO_KEY_MSG)
        # TODO: check interaction with WATCH
        if newkey.key != key.key:
            newkey.value = key.value
            newkey.expireat = key.expireat
            key.value = None
        return OK

    @command(name="RENAMENX", fixed=(Key(), Key()))
    def renamenx(self, key: CommandItem, newkey: CommandItem) -> int:
        if not key:
            raise SimpleError(msgs.NO_KEY_MSG)
        if newkey:
            return 0
        self.rename(key, newkey)
        return 1

    @command(name="RESTORE", fixed=(Key(), Int, bytes), repeat=(bytes,))
    def restore(self, key: CommandItem, ttl: int, value: bytes, *args: bytes) -> SimpleString:
        (replace,), _ = extract_args(args, ("replace",))
        if key and not replace:
            raise SimpleError(msgs.RESTORE_KEY_EXISTS)
        checksum, value = value[:20], value[20:]
        if hashlib.sha1(value).digest() != checksum:
            raise SimpleError(msgs.RESTORE_INVALID_CHECKSUM_MSG)
        if ttl < 0:
            raise SimpleError(msgs.RESTORE_INVALID_TTL_MSG)
        if ttl == 0:
            expireat = None
        else:
            expireat = self._db.time + ttl / 1000.0
        key.value = pickle.loads(value)
        key.expireat = expireat
        return OK

    @command(name="SCAN", fixed=(Int,), repeat=(bytes, bytes))
    def scan(self, cursor: int, *args: bytes) -> List[Union[bytes, List[bytes]]]:
        return self._scan(list(self._db), cursor, *args)

    @command(name="SORT", fixed=(Key(),), repeat=(bytes,))
    def sort(self, key: CommandItem, *args: bytes) -> Union[int, List[Any]]:
        if key.value is not None and not isinstance(key.value, (ExpiringMembersSet, list, ZSet)):
            raise SimpleError(msgs.WRONGTYPE_MSG)
        ((asc, desc, alpha, store, sortby, (limit_start, limit_count)), left_args) = extract_args(
            args,
            ("asc", "desc", "alpha", "*store", "*by", "++limit"),
            error_on_unexpected=False,
            left_from_first_unexpected=False,
        )
        limit_start = limit_start or 0
        limit_count = -1 if limit_count is None else limit_count
        dontsort = sortby is not None and b"*" not in sortby

        i = 0
        get = []
        while i < len(left_args):
            if casematch(left_args[i], b"get") and i + 1 < len(left_args):
                get.append(left_args[i + 1])
                i += 2
            else:
                raise SimpleError(msgs.SYNTAX_ERROR_MSG)

        # TODO: force sorting if the object is a set and either in Lua or storing to a key, to match redis behaviour.
        items = list(key.value) if key.value is not None else []

        # These transformations are based on the redis implementation, but changed to produce a half-open range.
        start = max(limit_start, 0)
        end = len(items) if limit_count < 0 else start + limit_count
        if start >= len(items):
            start = end = len(items) - 1
        end = min(end, len(items))

        if not get:
            get.append(b"#")
        if sortby is None:
            sortby = b"#"

        if not dontsort:

            def sort_key(val: bytes) -> Union[bytes, BeforeAny]:
                byval = self._lookup_key(val, sortby)
                # TODO: use locale.strxfrm when not storing? But then need to decode too.
                if byval is None:
                    return BeforeAny()
                return byval

            def sort_key_score(val: bytes) -> Tuple[float, bytes]:
                byval = self._lookup_key(val, sortby)
                score = SortFloat.decode(byval) if byval is not None else 0.0
                return score, val

            sort_func = sort_key if alpha else sort_key_score
            items.sort(key=sort_func, reverse=desc)
        # A `BY` pattern with no `*` means "don't sort": keep natural order (insertion order for lists, score order for
        # zsets) and only reverse when DESC is given.
        elif desc and isinstance(key.value, (list, ZSet)):
            items.reverse()

        out = []
        for row in items[start:end]:
            for g in get:
                v = self._lookup_key(row, g)
                if store is not None and v is None:
                    v = b""
                out.append(v)
        if store is not None:
            item = CommandItem(store, self._db, item=self._db.get(store))
            item.value = out
            item.writeback()
            return len(out)
        else:
            return out

    @command(name="SORT_RO", fixed=(Key(),), repeat=(bytes,))
    def sort_ro(self, key: CommandItem, *args: bytes) -> List[bytes]:
        if key.value is not None and not isinstance(key.value, (set, list, ZSet)):
            raise SimpleError(msgs.WRONGTYPE_MSG)
        ((asc, desc, alpha, sortby, (limit_start, limit_count)), left_args) = extract_args(
            args,
            ("asc", "desc", "alpha", "*by", "++limit"),
            error_on_unexpected=False,
            left_from_first_unexpected=False,
        )
        limit_start = limit_start or 0
        limit_count = -1 if limit_count is None else limit_count
        dontsort = sortby is not None and b"*" not in sortby

        i = 0
        get = []
        while i < len(left_args):
            if casematch(left_args[i], b"get") and i + 1 < len(left_args):
                get.append(left_args[i + 1])
                i += 2
            else:
                raise SimpleError(msgs.SYNTAX_ERROR_MSG)

        # TODO: force sorting if the object is a set and either in Lua or storing to a key, to match redis behaviour.
        items = list(key.value) if key.value is not None else []

        # These transformations are based on the redis implementation, but changed to produce a half-open range.
        start = max(limit_start, 0)
        end = len(items) if limit_count < 0 else start + limit_count
        if start >= len(items):
            start = end = len(items) - 1
        end = min(end, len(items))

        if not get:
            get.append(b"#")
        if sortby is None:
            sortby = b"#"

        if not dontsort:

            def sort_key(val: bytes) -> Union[bytes, BeforeAny]:
                byval = self._lookup_key(val, sortby)
                # TODO: use locale.strxfrm when not storing? But then need to decode too.
                if byval is None:
                    return BeforeAny()
                return byval

            def sort_key_score(val: bytes) -> Tuple[float, bytes]:
                byval = self._lookup_key(val, sortby)
                score = SortFloat.decode(byval) if byval is not None else 0.0
                return score, val

            sort_func = sort_key if alpha else sort_key_score
            items.sort(key=sort_func, reverse=desc)
        # A `BY` pattern with no `*` means "don't sort": keep natural order
        # (insertion order for lists, score order for zsets) and only reverse
        # when DESC is given.
        elif desc and isinstance(key.value, (list, ZSet)):
            items.reverse()

        out: List[bytes] = []
        for row in items[start:end]:
            for g in get:
                v = self._lookup_key(row, g)
                out.append(v)  # type:ignore
        return out

    @command(name="TTL", fixed=(Key(),))
    def ttl(self, key: CommandItem) -> int:
        return self._ttl(key, 1.0)

    @command(name="TYPE", fixed=(Key(),))
    def type_cmd(self, key: CommandItem) -> SimpleString:
        return self._key_value_type(key)

    @command(name="UNLINK", fixed=(Key(),), repeat=(Key(),))
    def unlink(self, *keys: CommandItem) -> int:
        return delete_keys(*keys)

    @command(name="COPY", fixed=(Key(), Key()), repeat=(bytes,))
    def copy(self, key: CommandItem, newkey: CommandItem, *args: bytes) -> int:
        (db_num, replace), _ = extract_args(args, ("+db", "replace"))
        db_num = db_num or self._db_num
        if key.key == newkey.key and db_num == self._db_num:
            raise SimpleError(msgs.SRC_DST_SAME_MSG)
        if (newkey.key in self._server.dbs[db_num] and not replace) or (key.key not in self._server.dbs[self._db_num]):
            return 0

        newkey.value = key.value
        newkey.expireat = key.expireat
        self._server.dbs[db_num][newkey.key] = key
        newkey.db = self._server.dbs[db_num]
        return 1



import re
from typing import Dict, List, Optional, Tuple


from fakeredis.model._base_type import BaseModel


class Array(BaseModel):
    _model_type = b"array"

    def __init__(self) -> None:
        self._data: Dict[int, bytes] = {}
        self._cursor: int = 0
        # ordered dict used as ordered set: index -> None, keyed by last-insert time
        self._insertion_order: Dict[int, None] = {}

    def __len__(self) -> int:
        return len(self._data)

    def length(self) -> int:
        """ARLEN: max_index + 1, or 0 if empty."""
        if not self._data:
            return 0
        return max(self._data) + 1

    def count(self) -> int:
        """ARCOUNT: number of non-empty (set) elements."""
        return len(self._data)

    def get(self, index: int) -> Optional[bytes]:
        return self._data.get(index)

    def set(self, index: int, value: bytes) -> bool:
        """Set value at index. Returns True if slot was previously empty."""
        is_new = index not in self._data
        self._data[index] = value
        return is_new

    def delete(self, index: int) -> bool:
        """Delete element at index. Returns True if something was deleted."""
        if index in self._data:
            del self._data[index]
            self._insertion_order.pop(index, None)
            return True
        return False

    def truncate_at(self, size: int) -> None:
        """Remove all elements at index >= size (used by ARRING)."""
        for k in [k for k in self._data if k >= size]:
            self.delete(k)

    def record_insert(self, index: int) -> None:
        """Record an ARINSERT/ARRING insertion for ARLASTITEMS tracking."""
        self._insertion_order.pop(index, None)
        self._insertion_order[index] = None

    def lastitems(self, count: int) -> List[bytes]:
        """Return the last `count` inserted values (oldest-first)."""
        existing = [i for i in self._insertion_order if i in self._data]
        recent = existing[-count:] if count < len(existing) else existing
        return [self._data[i] for i in recent]

    def scan_range(self, start: int, end: int, limit: Optional[int] = None) -> List[Tuple[int, bytes]]:
        """Return existing index-value pairs in [start, end] (inclusive).

        If start > end the range is traversed in descending order.
        Stops after `limit` pairs if given.
        """
        if start <= end:
            indices = sorted(k for k in self._data if start <= k <= end)
        else:
            indices = sorted((k for k in self._data if end <= k <= start), reverse=True)
        if limit is not None:
            indices = indices[:limit]
        return [(i, self._data[i]) for i in indices]

    def grep_range(
        self,
        start: int,
        end: int,
        predicates: List[Tuple[str, str]],
        use_and: bool,
        limit: Optional[int],
        nocase: bool,
    ) -> List[Tuple[int, bytes]]:
        """Return (index, value) pairs in range matching the textual predicates."""
        if start <= end:
            indices = sorted(k for k in self._data if start <= k <= end)
        else:
            indices = sorted((k for k in self._data if end <= k <= start), reverse=True)

        results: List[Tuple[int, bytes]] = []
        for idx in indices:
            val = self._data[idx]
            text = val.decode(errors="replace")
            if nocase:
                text = text.lower()

            matches = []
            for kind, pattern in predicates:
                p = pattern.lower() if nocase else pattern
                if kind == "exact":
                    matches.append(text == p)
                elif kind == "match":
                    matches.append(p in text)
                elif kind == "glob":
                    matches.append(_glob_match(p, text))
                elif kind == "re":
                    flags = re.IGNORECASE if nocase else 0
                    matches.append(bool(re.search(pattern, val.decode(errors="replace"), flags)))

            if not matches:
                continue
            hit = all(matches) if use_and else any(matches)
            if hit:
                results.append((idx, val))
                if limit is not None and len(results) >= limit:
                    break
        return results

    def op_range(self, start: int, end: int, operation: str, operand: Optional[bytes] = None):
        """Perform an aggregate operation on elements in [start, end]."""
        if start > end:
            start, end = end, start
        values = [self._data[k] for k in self._data if start <= k <= end]

        if operation == "used":
            return len(values)
        if operation == "match":
            return sum(1 for v in values if v == operand)

        if not values:
            return None

        if operation in ("and", "or", "xor"):
            result = 0
            for v in values:
                try:
                    n = int(float(v))
                except (ValueError, OverflowError):
                    n = 0
                if operation == "and":
                    result = result & n if result != 0 or values.index(v) > 0 else n
                elif operation == "or":
                    result |= n
                elif operation == "xor":
                    result ^= n
            return result

        # SUM, MIN, MAX
        nums = []
        for v in values:
            try:
                nums.append(float(v))
            except (ValueError, OverflowError):
                pass
        if not nums:
            return None
        if operation == "sum":
            total = sum(nums)
            if total == int(total):
                return str(int(total)).encode()
            return str(total).encode()
        if operation == "min":
            m = min(nums)
            if m == int(m):
                return str(int(m)).encode()
            return str(m).encode()
        if operation == "max":
            m = max(nums)
            if m == int(m):
                return str(int(m)).encode()
            return str(m).encode()
        return None


def _glob_match(pattern: str, text: str) -> bool:
    """Translate a glob pattern to regex and match."""
    regex = re.escape(pattern).replace(r"\*", ".*").replace(r"\?", ".").replace(r"\[", "[").replace(r"\]", "]")
    return bool(re.fullmatch(regex, text))



import logging
import threading
import time
import weakref
from collections import defaultdict
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, Type, Union

import redis

from fakeredis._helpers import Database, FakeSelector
from fakeredis._typing import ServerType, VersionType
from fakeredis.model import AccessControlList, ClientInfo

LOGGER = logging.getLogger("fakeredis")


def _create_version(v: Union[Tuple[int, ...], int, str]) -> VersionType:
    if isinstance(v, tuple):
        return v
    if isinstance(v, int):
        return (v,)
    if isinstance(v, str):
        v_split = v.split(".")
        return tuple(int(x) for x in v_split)
    raise ValueError(f"Unsupported version: {v}")


def _version_to_str(v: VersionType) -> str:
    if isinstance(v, tuple):
        return ".".join(str(x) for x in v)
    return str(v)


class FakeServer:
    _servers_map: Dict[str, "FakeServer"] = {}

    def __init__(
        self,
        version: VersionType = (8,),
        server_type: ServerType = "redis",
        config: Optional[Dict[bytes, bytes]] = None,
    ) -> None:
        """Initialize a new FakeServer instance.
        :param version: The version of the server (e.g. 6, 7.4, "7.4.1", can also be a tuple)
        :param server_type: The type of server (redis, dragonfly, valkey)
        :param config: A dictionary of configuration options.

        Configuration options:
        - `requirepass`: The password required to authenticate to the server.
        - `aclfile`: The path to the ACL file.
        """
        self.lock = threading.Lock()
        self.dbs: Dict[int, Database] = defaultdict(lambda: Database(self.lock))
        # Maps channel/pattern to a weak set of sockets
        self.script_cache: Dict[bytes, bytes] = {}  # Maps SHA1 to the script source
        self.subscribers: Dict[bytes, weakref.WeakSet[Any]] = defaultdict(weakref.WeakSet)
        self.psubscribers: Dict[bytes, weakref.WeakSet[Any]] = defaultdict(weakref.WeakSet)
        self.ssubscribers: Dict[bytes, weakref.WeakSet[Any]] = defaultdict(weakref.WeakSet)
        self.lastsave: int = int(time.time())
        self.connected = True
        # List of weakrefs to sockets that are being closed lazily
        self.sockets: List[Any] = []
        self.closed_sockets: List[Any] = []
        self.version: VersionType = _create_version(version)
        if server_type not in ("redis", "dragonfly", "valkey"):
            raise ValueError(f"Unsupported server type: {server_type}")
        self.server_type: ServerType = server_type
        self.config: Dict[bytes, bytes] = config or {}
        self.acl: AccessControlList = AccessControlList()
        self.clients: Dict[str, Dict[str, Any]] = {}
        self._next_client_id = 1

    def get_next_client_id(self) -> int:
        with self.lock:
            client_id = self._next_client_id
            self._next_client_id += 1
        return client_id

    @staticmethod
    def get_server(key: str, version: VersionType, server_type: ServerType) -> "FakeServer":
        if key not in FakeServer._servers_map:
            FakeServer._servers_map[key] = FakeServer(version=version, server_type=server_type)
        return FakeServer._servers_map[key]


class FakeBaseConnectionMixin(object):
    def __init__(
        self,
        *args: Any,
        version: VersionType = (7, 0),
        server_type: ServerType = "redis",
        server: Optional[FakeServer] = None,
        client_class: Type[redis.Redis] = redis.Redis,
        lua_modules: Optional[Set[str]] = None,
        writer: Any = None,
        connected: bool = True,
        **kwargs: Any,
    ) -> None:
        """
        Initializes the class and sets up the required attributes and configurations for the server and client interaction.

        """
        self.client_name: Optional[str] = None
        self.server_key: str
        self._sock = None
        self._selector: Optional[FakeSelector] = None
        self._server = server
        self._client_class = client_class
        self._lua_modules = lua_modules
        self._writer = writer
        if self._server is None:
            if "path" in kwargs:
                self.server_key = kwargs.pop("path")
            else:
                host, port = kwargs.get("host"), kwargs.get("port")
                self.server_key = f"{host}:{port}"
            self.server_key += f":{server_type}:v{_version_to_str(version)[0]}"
            self._server = FakeServer.get_server(self.server_key, server_type=server_type, version=version)
            self._server.connected = connected
        client_info_arg = kwargs.pop("client_info", {})
        super().__init__(*args, **kwargs)
        protocol = getattr(self, "protocol", 2)

        client_info = dict(
            id=self._server.get_next_client_id(),
            addr="127.0.0.1:0",
            laddr="127.0.0.1:6379",
            fd=8,
            name="",
            idle=0,
            flags="N",
            db=0,
            sub=0,
            psub=0,
            ssub=0,
            multi=-1,
            qbuf=48,
            qbuf_free=16842,
            argv_mem=25,
            multi_mem=0,
            rbs=1024,
            rbp=0,
            obl=0,
            oll=0,
            omem=0,
            tot_mem=18737,
            events="r",
            cmd="auth",
            redir=-1,
            resp=protocol,
        )
        client_info.update(client_info_arg)
        self._client_info = ClientInfo(**client_info)

    def _decode(self, response: Any) -> Any:
        if isinstance(response, list):
            return [self._decode(item) for item in response]
        elif isinstance(response, dict):
            return {self._decode(k): self._decode(v) for k, v in response.items()}
        elif isinstance(response, bytes):
            return self.encoder.decode(response)  # type: ignore[attr-defined]
        else:
            return response

    def _add_to_local_cache(self, command: Sequence[str], response: Any, keys: List[Any]) -> None:
        return None

    def repr_pieces(self) -> List[Tuple[str, Any]]:
        pieces = [("server", self._server), ("db", self.db)]  # type: ignore[attr-defined]
        if self.client_name:
            pieces.append(("client_name", self.client_name))
        return pieces

    def __str__(self) -> str:
        return self.server_key
